import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from scipy.signal import savgol_filter
import logging

from .flare_checks import check_noise, check_slopes, check_above, check_variance

logger = logging.getLogger('laff')

def flares_savgol(data, **kwargs) -> list:

    """_summary_

    Args:
        data (pd.DataFrame): _description_
        algorithm (_type_): _description_

        plot_savgol(bool): overlay the filter used on an ongoing plt object.
    Returns:
        _type_: _description_
    """


    logger.debug("Starting sequential_findflares()")

    if len(data.index) > 15:
        size = 15
    else: size = 4
    
    data['savgol'] = savgol_filter(data.flux, window_length=size, polyorder=3)

    if 'plot_savgol' in kwargs:
        plt.plot(data['time'], data['savgol'], color='m', linestyle='--', linewidth=0.5)
    
    final_index = len(data.flux) - 2
    
    n = 0
    prev_decay = 0
    FLARES = []

    while n < final_index:

        search_start = n
        search_count = 0
        
        # Run deviation check.
        if data.iloc[n+1].savgol > data.iloc[n].savgol:
            search_count = 1

            # Boundary if we reach end of data.
            # print('nsearch2', n+search_count+2)
            # print('finalindex', final_index)
            if n+search_count+2 >= final_index:
                n = final_index
                continue

            try:
                while data.iloc[n+search_count+1].savgol >= data.iloc[n+search_count].savgol:
                    search_count += 1
            except IndexError:

                # reach end of data
                search_count = 0
                n = final_index
                logger.debug('Reached end of data, ending flare search.')
                continue

        if search_count >= 2:

            logger.debug("Possible deviation from %s -> %s (t=%s)", search_start, search_start+search_count, data.iloc[search_start].time)

            start_point = find_start(data, search_start, prev_decay)
            peak_point = find_peak(data, start_point)

            ## Quick rise check.
            savgol_rise_start = data['savgol'].iloc[peak_point] > data['savgol'].iloc[start_point] + 2*(data['flux_perr'].iloc[start_point])
            savgol_rise_before = data['savgol'].iloc[peak_point] > np.average(data['savgol'].iloc[max(start_point - 5, 0):start_point] + 2*(data['flux_perr'].iloc[max(start_point - 5, 0):start_point]))

            if not savgol_rise_start and not savgol_rise_before:
                logger.debug("Doesn't meet savgol rise condition")
                n += 1
                continue

            peak_point, decay_point = find_decay(data, start_point, peak_point)

            checks = [check_noise(data, start_point, peak_point),
                      check_slopes(data, start_point, peak_point, decay_point),
                      check_above(data, start_point, decay_point)]
                    #   check_variance(data, peak_point, decay_point)]    
            logger.debug(f"\tChecks: {checks}")

            if all(checks):
                check_variance(data, peak_point, decay_point)
                check_noise(data, start_point, peak_point)
                FLARES.append([start_point, peak_point, decay_point])
                logger.debug(f"\tconfirmed flare::  {start_point, peak_point, decay_point}")
                n = decay_point
                prev_decay = decay_point
                continue
            else:
                # All checks failed.
                logger.debug("Flare failed passing all tests - discarding.")
        else:
            # search_count not greater than 2, move on.
            pass

        n += 1

    return FLARES


def find_start(data: pd.DataFrame, start: int, prev_decay: int) -> int:
    """Return flare start by looking for local minima."""


    ## dev - test
    # with the savgol smoothing, somestimes the wrong point is selected.
    if data['flux'].iloc[start + 1] < data['flux'].iloc[start]:

        bottom_next = data['flux'].iloc[start + 1] + data['flux_nerr'].iloc[start + 1]
        top_prev    = data['flux'].iloc[start - 1] + data['flux_perr'].iloc[start - 1]

        if bottom_next < top_prev:
            minimum = start + 1
            logger.debug("\tflare start found at %s (t=%s)", minimum, data['time'].iloc[minimum])
            return minimum

    if start < 3:
        points = data.iloc[0:3]
    else:
        points = data.iloc[start-3:start+2]
    minimum = data[data.flux == min(points.flux)].index.values[0]
    minimum = prev_decay if (minimum < prev_decay) else minimum

    logger.debug(f"\tflare start found at {minimum}")
    return minimum


def find_peak(data, start):
    """
    Return flare peak by looking for local maxima.

    Starting at point {start}, look for the peak of the flare. Since this could
    be one, or many points away a moving average algorithm is used. Work out
    the average of 5 point chunks and see if this is still rising. Until the
    rise stops, continue to search. Once a decay has been found, the peak is the
    datapoint with maximum value.

    :param data: The pandas dataframe containing the lightcurve data.
    :param start: Integer position of the flare start.
    :return: Integer position of the flare peak.
    """

    chunksize = 4
    prev_chunk = data['flux'].iloc[start] # Flare start position is first point.
    next_chunk = np.average(data.iloc[start+1:start+1+chunksize].flux) # Calculate first 'next chunk'.

    # boundary for end of data
    if start + 1 + chunksize >= len(data.index): # has looped around
        points = data.iloc[start+1:len(data.index)]
        maximum = data[data.flux == max(points.flux)].index.values[0]

        logger.debug(f"\tFlare peak found at {maximum} - using end of data cutoff.")
        return maximum

    i = 1

    while next_chunk > prev_chunk:
        # Next chunk interation.
        i += 1
        prev_chunk = next_chunk
        next_chunk = np.average(data.iloc[(start+1)+(chunksize*i):(start+1+chunksize)+(chunksize*i)].flux)
    else:
        # Data has now begin to descend so look for peak up to these points.
        # Include next_chunk in case the peak lays in this list, but is just
        # brought down as an average by remaining points.
        points = data.iloc[start:(start+1+chunksize)+(chunksize*i)]
        maximum = data[data.flux == max(points.flux)].index.values[0]

    logger.debug("\tFlare peak found at %s (t=%s)", maximum, data['time'].iloc[maximum])
    return maximum


def find_decay(data: pd.DataFrame, start: int, peak: int) -> int:
    """
    Find the end of the flare as the decay smoothes into continuum.

    Longer description.

    :param data:
    :param peak:
    :returns:
    """

    decay = peak
    condition = 0
    # decaypar = 2.5

    logger.debug("\tlooking for decay")

    def calc_grad(data: pd.DataFrame, idx1: int, idx2: int, peak: bool = False) -> int:
        """Calculate gradient between first (idx1) and second (idx2) points."""
        deltaFlux = data.iloc[idx2].savgol - data.iloc[idx1].flux if peak else data.iloc[idx2].savgol - data.iloc[idx1].savgol
        deltaTime = data.iloc[idx2].time - data.iloc[idx1].time
        if deltaTime < 0:
            raise ValueError("It appears data is not sorted in time order. Please fix this.")
        return deltaFlux/deltaTime

    while condition < 3:
        decay += 1

        # Boundary condition for end of data.
        if data.idxmax('index').time in [decay + i for i in range(-1,2)]:  # reach end of data
            logger.debug(f"Reached end of data, automatically ending flare at {decay + 1}")
            condition = 3
            decay = data.idxmax('index').time
            continue

        # Condition for large orbital gaps.
        if (data['time'].iloc[decay+1] - data['time'].iloc[decay]) > (data['time'].iloc[decay] - data['time'].iloc[peak]) * 3:
            logger.debug(f"Gap between {decay}->{decay+1} is greater than {peak}->{decay} * 3")
            condition = 3
            continue

        # Prevent early ending - should decay to below start.
        if data['flux'].iloc[decay] > data['flux'].iloc[start]:
            condition = 0 # should this always be the case?
            continue

        # Calculate gradients.
        NextAlong = calc_grad(data, decay, decay+1)
        PrevAlong = calc_grad(data, decay-1, decay)
        PeakToCurrent = calc_grad(data, peak, decay, peak=True)
        PeakToPrev = calc_grad(data, peak, decay-1, peak=True)
        
        # print(NextAlong, PrevAlong, PeakToCurrent, PeakToPrev)

        cond1 = NextAlong > PeakToCurrent # Next sequence is shallower than from peak to next current.
        cond2 = NextAlong > PrevAlong # Next grad is shallower than previous grad.
        cond3 = PeakToCurrent > PeakToPrev # Peak to next point is shallower than from peak to previous point.

        if cond1 and cond2 and peak == decay - 1: # special case for first test only
            cond3 = True

        # Evaluate conditions - logic in notebook 20th august.
        # if cond1 and cond2 and cond3:
        #     condition += 1
        # else:
        #     condition = condition - 1 if condition > 0 else 0

        ## dev
        # print(f"At {decay} conditions are [{cond1, cond2, cond3}] and condition before eval is {condition}")
        ## dev


        if cond1 and cond2 and cond3:
            if condition == 2:
                condition = 3
            elif condition == 1:
                condition = 3
            elif condition == 0:
                condition = 2
        else:
            if condition == 2:
                condition = 1
            if condition == 1:
                condition = 0

        if (data['savgol'].iloc[decay] > data['flux'].iloc[start]):
            condition = 0

    logger.debug("\tdecay found at %s (t=%s)", decay, data['time'].iloc[decay])

    # Adjust end for local minima.
    decay = data[data.flux == min(data.iloc[decay-1:decay+1].flux)].index.values[0]
    if decay <= peak:
        logger.debug("decay == peak, adding one")
        # raise ValueError('decay is before or on the peak')
        decay = peak + 1

    # Check flare peak adjustments.
    adjusted_flare_peak = data[data.flux == max(data.iloc[peak:decay].flux)].index.values[0]

    if peak < adjusted_flare_peak:
        logger.debug(f"Flare peak adjust from {peak} to {adjusted_flare_peak} - likely a noisy/multiple flare.")
        peak = adjusted_flare_peak

    return peak, decay


    # once end is found we will check if the flare is 'good'
    # if flare is good, accept it and continue search -> from end + 1
    # if flare is not good, disregard and continue search from deviation + 1



import logging
import pandas as pd
import numpy as np

##
import matplotlib.pyplot as plt
##

logger = logging.getLogger('laff')


def check_noise(data: pd.DataFrame, start: int, peak: int) -> bool:
    """Flare rise must be greater than 4x the local noise."""

    flare_rise = data['flux'].iloc[peak] - data['flux'].iloc[start]
    noise_level = np.average(data['flux_perr'].iloc[start:peak])

    return flare_rise > 3 * noise_level

def check_slopes(data: pd.DataFrame, start: int, peak: int, decay: int) -> bool:
    """Check the fraction of increases/decreases during the rise and de cay, respectively."""

    increase_threshold = 0.60
    decrease_threshold = 0.65

    rise_flux = data['flux'].iloc[start:peak+1]
    increase_fraction = np.sum(np.diff(rise_flux) > 0) / len(rise_flux)

    # decay_flux = data['flux'].iloc[peak:decay+1]
    # decrease_fraction = np.sum(np.diff(decay_flux) < 0) / len(decay_flux)

    return True
    return increase_fraction >= increase_threshold # `and decrease_fraction > decrease_threshold


def check_above(data: pd.DataFrame, start: int, decay: int) -> bool:
    """
    Check the flare is above the (estimated) continuum.
    
    We calculate the powerlaw continuum through the flare by solving a set of
    power functions for (x, y) co-ordinates corresponding to the found flare
    start and end times. The number of points above and below the slope can then
    be found. If the fraction above the continuum is below 0.7, to allow some
    variation through noise, we discard the flare.

    """

    above_threshold = 0.8

    x_coords = data['time'].iloc[start], data['time'].iloc[decay]
    y_coords = data['flux'].iloc[start], data['flux'].iloc[decay]

    ## Solving y = nx^a for start and stop.
    alpha = np.emath.logn(x_coords[1]/x_coords[0], y_coords[1]/y_coords[0])
    norm = y_coords[1] / x_coords[1] ** alpha
    
    points_above = sum(flux > (norm*time**alpha) for flux, time in zip(data['flux'].iloc[start:decay], data['time'].iloc[start:decay]))
    num_points = len(data['flux'].iloc[start:decay])

    logger.debug(f"\tpoints above/num_points => {points_above}/{num_points} = {points_above/num_points}")

    return points_above/num_points > above_threshold




# def check_rise(data: pd.DataFrame, start: int, peak: int) -> bool:
#     """Check the index of the rise, and that no points drop below flare start during the rise."""

#     # During rise, no point should be below the start.
#     start_flux = data['flux'].iloc[start]
#     rise_fluxes = data['flux'].iloc[start+1:peak]
#     if any(value < start_flux for value in rise_fluxes):
#         return False


def check_variance(data: pd.DataFrame, peak: int, decay: int) -> bool:
    """Check the variance of the decay phase."""

    variance_threshold = 0.75

    decay_flux = data['flux'].iloc[peak:decay+1]
    peak_flux  = data['flux'].iloc[peak]
    variance = np.var(decay_flux) / peak_flux

    return variance < variance_threshold


import logging
import pandas as pd
import numpy as np

logger = logging.getLogger('laff')

def sequential(data, smooth=False) -> list:
    logger.debug("Starting sequential_findflares()")

    data_original = data.copy()

    # Apply smoothing function.
    if smooth == True:
        data['smoothed_flux'] = 10**(np.log10(data['flux']).rolling(window=3).mean())
        data.at[0, 'smoothed_flux'] = data['flux'].iloc[0]

    #######

    final_index = len(data.flux) - 2
    n = 0
    prev_decay = 0

    FLARES = []

    while n < final_index:

        logger.debug(f"Looking at index {n}")

        search_start = n
        search_count = 0
        
        # Run deviation check.
        if data.iloc[n+1].flux > data.iloc[n].flux:
            search_count = 1

            # Boundary if we reach end of data.
            if n+search_count+1 >= final_index:
                n = final_index
                continue

            while data.iloc[n+search_count+1].flux >= data.iloc[n+search_count].flux:
                search_count += 1

        if search_count >= 2:
            logger.debug(f"Possible deviation from {search_start}->{search_start+search_count} ({data.iloc[search_start].time})")

            start_point = find_start(data, search_start, prev_decay)
            peak_point = find_peak(data, start_point)

            logger.debug(f'Possible flare rise from {start_point}->{peak_point}')

            if check_rise(data, start_point, peak_point):
                peak_point, decay_point = find_decay(data, peak_point)
                checks = [check_noise(data, start_point, peak_point, decay_point),
                          check_above(data, start_point, decay_point),
                          check_decay_shape(data, peak_point, decay_point)]
                #dev
                # checks = [True for x in checks]
                #dev
                logger.debug(f"Checks: {checks}")

                if all(checks):
                    FLARES.append([start_point, peak_point, decay_point])
                    logger.debug(f"Confirmed flare::  {start_point, peak_point, decay_point}")
                    n = decay_point
                    prev_decay = decay_point
                    continue
                else:
                    # All checks failed.
                    logger.debug("Flare failed passing all tests - discarding.")
            else:
                # Check failed.
                logger.debug(f"Deviations has NOT passed checks - discarding")
        else:
            # search_count not greater than 2, move on.
            pass

        n += 1

    # Reset data.
    data = data_original.copy()
    return FLARES


def find_start(data: pd.DataFrame, start: int, prev_decay: int) -> int:
    """Return flare start by looking for local minima."""
    if start < 3:
        points = data.iloc[0:3]
    else:
        points = data.iloc[start-3:start+1]
    minimum = data[data.flux == min(points.flux)].index.values[0]
    minimum = prev_decay if (minimum < prev_decay) else minimum
    logger.debug(f"Flare start found at {minimum}")

    return minimum


def find_peak(data, start):
    """
    Return flare peak by looking for local maxima.

    Starting at point {start}, look for the peak of the flare. Since this could
    be one, or many points away a moving average algorithm is used. Work out
    the average of 5 point chunks and see if this is still rising. Until the
    rise stops, continue to search. Once a decay has been found, the peak is the
    datapoint with maximum value.

    :param data: The pandas dataframe containing the lightcurve data.
    :param start: Integer position of the flare start.
    :return: Integer position of the flare peak.
    """

    chunksize = 4
    prev_chunk = data['flux'].iloc[start] # Flare start position is first point.
    next_chunk = np.average(data.iloc[start+1:start+1+chunksize].flux) # Calculate first 'next chunk'.
    i = 1

    while next_chunk > prev_chunk:
        logger.debug(f"Looking at chunk i={i} : {(start+1)+(chunksize*i)}->{(start+1+4)+(chunksize*i)}")
        # Next chunk interation.
        i += 1
        prev_chunk = next_chunk
        next_chunk = np.average(data.iloc[(start+1)+(chunksize*i):(start+1+chunksize)+(chunksize*i)].flux)
    else:
        # Data has now begin to descend so look for peak up to these points.
        # Include next_chunk in case the peak lays in this list, but is just
        # brought down as an average by remaining points.
        points = data.iloc[start:(start+1+chunksize)+(chunksize*i)]
        maximum = data[data.flux == max(points.flux)].index.values[0]

        logger.debug(f"Flare peak found at {maximum}")
    return maximum


def find_decay(data: pd.DataFrame, peak: int) -> int:
    """
    Find the end of the flare as the decay smoothes into continuum.

    Longer description.

    :param data:
    :param peak:
    :returns:
    """

    # Apply smoothing function.
    data_smth = data.copy(deep=True)
    data_smth['flux'] = 10**(np.log10(data_smth['flux']).rolling(window=2).mean())

    # data['smoothed_flux'] = 10**(np.log10(data['flux']).rolling(window=2).mean())
    # data.at[0, 'smoothed_flux'] = data['flux'].iloc[0]

    decay = peak
    condition = 0
    # decaypar = 2.5

    logger.debug(f"Looking for decay")

    def calc_grad(data: pd.DataFrame, idx1: int, idx2: int) -> int:
        """Calculate gradient between first (idx1) and second (idx2) points."""
        deltaFlux = data.iloc[idx2].flux - data.iloc[idx1].flux
        deltaTime = data.iloc[idx2].time - data.iloc[idx1].time
        return deltaFlux/deltaTime

    while condition < 3:
        decay += 1
        if data_smth.idxmax('index').time in [decay + i for i in range(-1,2)]:  # reach end of data
            logger.debug(f"Reached end of data, automatically ending flare at {decay + 1}")
            condition = 3
            decay = data_smth.idxmax('index').time
            continue
        if (data['time'].iloc[decay+1] - data['time'].iloc[decay]) > (data['time'].iloc[decay] - data['time'].iloc[peak]) * 10:
            logger.debug(f"Gap between {decay}->{decay+1} is greater than {peak}->{decay} * 10")
            condition = 3
            decay += 1
            continue

        # Calculate gradients.
        NextAlong = calc_grad(data_smth, decay, decay+1)
        PrevAlong = calc_grad(data_smth, decay-1, decay)
        PeakToCurrent = calc_grad(data_smth, peak, decay)
        PeakToPrev = calc_grad(data_smth, peak, decay-1)

        cond1 = NextAlong > PeakToCurrent # Next sequence is shallower than from peak to next current.
        cond2 = NextAlong > PrevAlong # Next grad is shallower than previous grad.
        cond3 = PeakToCurrent > PeakToPrev # Peak to next point is shallower than from peak to previous point.

        if cond1 and cond2 and peak == decay - 1: # special case for first test only
            cond3 = True

        # Evaluate conditions - logic in notebook 20th august.
        if cond1 and cond2 and cond3:
            condition += 1
        else:
            condition = condition - 1 if condition > 0 else 0

        # if cond1 and cond2 and cond3:
        #     if condition == 2:
        #         condition = 3
        #     elif condition == 1:
        #         condition = 3
        #     elif condition == 0:
        #         condition = 2
        # else:
        #     if condition == 2:
        #         condition = 1
        #     if condition == 1:
        #         condition = 0


    logger.debug(f"Decay end found at {decay - 1}")
    decay = decay - 1


    # Check flare peak adjustments.
    adjusted_flare_peak = data[data.flux == max(data.iloc[peak:decay].flux)].index.values[0]

    if peak < adjusted_flare_peak:
        logger.debug(f"Flare peak adjust from {peak} to {adjusted_flare_peak} - likely a noisy/multiple flare.")
        peak = adjusted_flare_peak
    elif peak > adjusted_flare_peak:
        raise ValueError("Maximum was somehow previously missed.")

    return peak, decay


    # once end is found we will check if the flare is 'good'
    # if flare is good, accept it and continue search -> from end + 1
    # if flare is not good, disregard and continue search from deviation + 1

def check_rise(data: pd.DataFrame, start: int, peak: int) -> bool:
    """Test the rise is significant enough."""
    if data.iloc[peak].flux > data.iloc[start].flux + (2 * data.iloc[start].flux_perr):
        logger.debug("check_rise: true")
        return True
    else:
        logger.debug("check_rise: false")
        return False


def check_noise(data: pd.DataFrame, start: int, peak: int, decay: int) -> bool:
    """Check if flare is greater than x1.75 the average noise across the flare."""
    average_noise = abs(np.average(data.iloc[start:decay].flux_perr)) + abs(np.average(data.iloc[start:decay].flux_nerr))
    flux_increase = data.iloc[peak].flux - data.iloc[start].flux
    logger.debug(f"noise: {average_noise} | delta_flux: {flux_increase}")
    return True if flux_increase > 1.75 * average_noise else False

# def check_shape(data: pd.DataFrame, start: int, peak:int, decay:int) -> bool:
    # """Check the shape of the flare."""

def check_above(data: pd.DataFrame, start: int, decay: int) -> bool:
    """
    Check the flare is above the (estimated) continuum.
    
    We calculate the powerlaw continuum through the flare by solving a set of
    power functions for (x, y) co-ordinates corresponding to the found flare
    start and end times. The number of points above and below the slope can then
    be found. If the fraction above the continuum is below 0.7, to allow some
    variation through noise, we discard the flare.

    """
    # Check flare boundaries.
    start = 0 if start == 0 else start - 1
    decay = data.idxmax('index').time if decay == data.idxmax('index').time else decay + 1

    x_coords = data['time'].iloc[start], data['time'].iloc[decay]
    y_coords = data['flux'].iloc[start], data['flux'].iloc[decay]

    ## Solving y = nx^a for start and stop.
    alpha = np.emath.logn(x_coords[1]/x_coords[0], y_coords[1]/y_coords[0])
    norm = y_coords[1] / x_coords[1] ** alpha
    # logger.debug(f"ALPHA IS {alpha}")
    # logger.debug(f"NORM IS {norm}")
    
    points_above = sum(flux > (norm*time**alpha) for flux, time in zip(data['flux'].iloc[start:decay], data['time'].iloc[start:decay]))
    num_points = len(data['flux'].iloc[start:decay])

    logger.debug(f"points above/num_points => {points_above}/{num_points} = {points_above/num_points}")

    # return True
    return True if points_above/num_points >= 0.7 else False

def check_decay_shape(data: pd.DataFrame, peak: int, decay: int):

    decay_data = list(data.iloc[peak:decay].flux_perr)
    count_decrease = sum(b < a for a, b in zip(decay_data, decay_data[1:]))

    decay_shape = count_decrease / len(decay_data)
    logger.debug(f"decay shape {decay_shape}")
    
    if len(decay_data) < 4 and decay_shape > 0.1:
        return True
    
    # print("DECAY SHAPE VALUE", decay_shape)
    return True if decay_shape >= 0.5 else False


import logging
import numpy as np
import pandas as pd

logger = logging.getLogger('laff')

def flare_finding(data, algorithm, **kwargs):
    """_summary_

    Args:
        data (_type_): _description_
        algorithm (_type_): _description_

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """

    ## Choose algorithm.
    if algorithm in ('default', 'savgol', ''):
        from .algorithms import flares_savgol
        flares = flares_savgol(data, **kwargs)
    elif algorithm == 'sequential':
        # oudated
        from .algorithms import sequential
        flares = sequential(data)
    else:
        raise ValueError("invalid algorithm used")
    
    ## Clause for no flares.
    if flares is not False:
        logger.info("Found %s flare(s).", len(flares))
    else:
        logger.info('No flares found')
        
    return flares




## Main functions.
from .laff import fitGRB, findFlares, fitFlares, fitAfterglow
from .laff import plotGRB, printGRB
from .lightcurve_import import lcimport
from .fit_bat import fitPrompt, plotPrompt, fred_flare

## Auxilliary.
from .laff import set_logging_level
from .utility import STAT_NAMES_CONTINUUM, PAR_NAMES_FLARE, PAR_NAMES_CONTINUUM


import numpy as np
import logging
from scipy.optimize import fmin_slsqp
from ..utility import calculate_fit_statistics, calculate_par_err, calculate_fluence

logger = logging.getLogger('laff')

#################################################################################
### AFTERGLOW MODEL
#################################################################################

def broken_powerlaw(params, x):
    x = np.array(x)

    if type(params) in (list, np.ndarray):
        n = int((len(params)-2)/2)
        slopes = params[0:n+1]
        breaks = params[n+1:-1]
        norm   = params[-1]
    elif type(params) is dict:
        n      = params['break_num']
        slopes = params['slopes']
        breaks = params['breaks']
        norm   = params['normal']
    else:
        logger.critical('Input params not accepted type.')
        raise TypeError(f'params is not dict/list -> {type(params)}')
    
    breaks = [10**val for val in breaks]

    mask = []

    for i in range(n):
        mask.append(x > breaks[i])

    if n >= 0:
        model = norm * (x**(-slopes[0]))
    if n >= 1 and mask[0].any():
        model[np.where(mask[0])] = norm * (x[np.where(mask[0])]**(-slopes[1])) * (breaks[0]**(-slopes[0]+slopes[1]))
    if n >= 2 and mask[1].any():
        model[np.where(mask[1])] = norm * (x[np.where(mask[1])]**(-slopes[2])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2]))
    if n >= 3 and mask[2].any():
        model[np.where(mask[2])] = norm * (x[np.where(mask[2])]**(-slopes[3])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2])) * (breaks[2]**(-slopes[2]+slopes[3]))
    if n >= 4 and mask[3].any():
        model[np.where(mask[3])] = norm * (x[np.where(mask[3])]**(-slopes[4])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2])) * (breaks[2]**(-slopes[2]+slopes[3])) * (breaks[3]**(-slopes[3]+slopes[4]))
    if n >= 5 and mask[4].any():
        model[np.where(mask[4])] = norm * (x[np.where(mask[4])]**(-slopes[5])) * (breaks[0]**(-slopes[0]+slopes[1])) * (breaks[1]**(-slopes[1]+slopes[2])) * (breaks[2]**(-slopes[2]+slopes[3])) * (breaks[3]**(-slopes[3]+slopes[4])) * (breaks[4]**(-slopes[4]+slopes[5]))

    return model

def sum_residuals(params, *args):
    x, y, y_err, _, _ = args
    return np.sum(((y - broken_powerlaw(params, x)) / y_err) ** 2)    

#################################################################################
### FITTING
#################################################################################

def find_afterglow_fit(data, data_flare):

    data_start, data_end = data['time'].iloc[0], data['time'].iloc[-1]

    model_fits = []

    logger.debug('breaknum : fit_par / fit_err / fit_stats')
    for breaknum in range(0, 6):

        # Guess parameters.
        slope_guesses = [1.0] * (breaknum+1)
        break_guesses = list(np.linspace(np.log10(data_start), np.log10(data_end), num=breaknum+2))[1:-1]
        normal_guess  = [data['flux'].iloc[0]]
        input_par = slope_guesses + break_guesses + normal_guess

        # Parameter bounds.
        if breaknum == 0:
            bounds = ([-0.3, 6.0], [0, np.inf])
        else:
            bounds = tuple([[-0.3, 6.0]] * (breaknum + 1) + [[np.log10(data['time'].iloc[0])+0.05, np.log10(data['time'].iloc[-1])-0.05]] * breaknum + [[0, np.inf]])
        
        # Constraints.
        def all_constraints(params, *args):
            _, _, _, flare_x, flare_y = args
            ordered_breaks = np.diff(params[breaknum+1:-1]) - 0.05 # order breaks
            flare_limits = flare_y - broken_powerlaw(params, flare_x) # flares as upper lims
            return np.concatenate([ordered_breaks, flare_limits])
        
        fit_par = fmin_slsqp(sum_residuals, input_par, bounds=bounds, f_ieqcons=all_constraints, args=(data.time, data.flux, data.flux_perr, data_flare[0], data_flare[1]), iter=500, iprint=0)
        fit_stats = calculate_fit_statistics(data, broken_powerlaw, fit_par)
        model_fits.append([fit_par, fit_stats])
        
        #####

    # Assess best fit.
    best_fit, best_stats = min(model_fits, key=lambda x: x[1]['deltaAIC'])
    breaknum = int((len(best_fit)-2)/2)

    def chi2_wrapper(params):
        return sum_residuals(params, data['time'], data['flux'], data['flux_perr'], 1, 1)
    
    param_errors = calculate_par_err(best_fit, chi2_wrapper)
    
    logger.info('Afterglow fitted with %s breaks.', breaknum)
    logger.debug('slopes\t%s', list(round(x,2) for x in best_fit[0:breaknum+1]))
    logger.debug('slopes_err\t%s', list(round(x,2) for x in param_errors[0:breaknum+1]))
    logger.debug('breaks\t%s', list(round(10**x,2) for x in best_fit[breaknum+1:-1]))
    logger.debug('breaks_err\t%s', list(round(10**x,2) for x in param_errors[breaknum+1:-1]))
    logger.debug('norm\t%s', best_fit[-1])
    logger.debug('norm\t%s', param_errors[-1])

    return list(best_fit), list(param_errors), best_stats, breaknum

#################################################################################
# CALCULATE FLUENCE
#################################################################################

def calculate_afterglow_fluence(data, breaknum, break_times, fit_par, count_ratio):

    break_times = [10**(x) for x in break_times]

    integral_boundaries = [data.iloc[0].time, *break_times, data.iloc[-1].time]

    phase_fluences = [calculate_fluence(broken_powerlaw, fit_par, integral_boundaries[i], integral_boundaries[i+1], count_ratio) * count_ratio for i in range(breaknum)]
    total_fluence = np.sum(phase_fluences)
    logger.info('Afterglow fluence calculated as %s', total_fluence)

    return total_fluence, phase_fluences


import pandas as pd
import matplotlib.pyplot as plt
import numpy as np
import logging
import warnings

# Ignore warnings.
# from pandas.core.common import SettingWithCopyWarning
# warnings.simplefilter(action='ignore', category=SettingWithCopyWarning)
warnings.filterwarnings("ignore", category=RuntimeWarning)

from .flarefinding import flare_finding
from .modelling import broken_powerlaw, find_afterglow_fit, calculate_afterglow_fluence
from .modelling import flare_fitter, package_flares, fred_flare
from .utility import check_data_input, calculate_fit_statistics, calculate_fluence

## recent note:
## the conditions in fit_afterglow work but the fitter gets 'stuck' at that
## value without changing input variables.
## either remove conditions and filter after or find a force shuffle variables.

# findFlares() -- locate the indices of flares in the lightcurve
# fitContinuum(flare_indices) -- use the indices to exclude data, then fit the continuum
# fitFlares(flares, continuum) -- use indices + continuum to fit the flares

# fitGRB() -- runs all 3 function in sequence, then does some final cleanups
#          -- final statistics of the whole fit
#          -- this is what the user should be running
#          -- outputs a dictionary with all useful statistics

################################################################################
### LOGGER
################################################################################

logging_level = 'INFO'
logger = logging.getLogger('laff')
logger.setLevel(logging_level)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)

def set_logging_level(level):
    """Set the desired logging level of the script.

    Args:
        level (string): from least to most verbose - 'none', 'debug', 'info',
        'warning', 'error', 'critical'. The default level is normal.

    Raises:
        ValueError: Invalid logging level.
    """

    if level.lower() in ['debug', 'info', 'warning', 'error', 'critical']:
        logging_level = level.upper()
        logger.setLevel(logging_level)
    elif level.lower() in ['verbose']:
        logger.setLevel('DEBUG')
    elif level.lower() in ['none', 'quiet']:
        logger.setLevel(60) # set to above all other levels
    else:
        raise ValueError("Invalid logging level. Please use 'DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL' or 'NONE'.")

################################################################################
### FIND FLARES
################################################################################

def findFlares(data, algorithm='savgol', **kwargs):
    """Identify flares within datasets."""

    logger.debug(f"Starting findFlares - method {algorithm}")
    if check_data_input(data) == False:
        return  # First check input format is good.

    # Run flare finding.
    flares = flare_finding(data, algorithm, **kwargs)

    return flares if len(flares) else False

################################################################################
### CONTINUUM FITTING
################################################################################

def fitAfterglow(data: pd.DataFrame, flare_indices: list[list[int]] = None, *, errors_to_std: float = 1.0, count_ratio: float = 1.0) -> dict:
    """Fits the afterglow of the light curve with a series of broken power laws.

    Args:
        data (pd.DataFrame):
            The dataset stored as a pandas dataframe.
        flare_indices (list[list[int]]):
            A nested list of 3 integers, the start/peak/end of flares as
            returned by laff.findFlares().
        errors_to_std (float, optional):
            The conversion factor to be applied to the x and y errors on data,
            the ODR fitter assumes there are 1-sigma standard deviations.
            Defaults to 1.0.
        count_flux_ratio (float, optional):
            The conversion factor to be applied to scale into flux, if the data
            provided is in count rate. Defaults to 1.0.

    Raises:
        ValueError: _description_
        ValueError: _description_

    Returns:
        dict: _description_
    """

    logger.debug('fitAfterglow()')

    # Separate flare data.
    data_flare = [[], []]
    if flare_indices:
        logger.debug('Removing indices of %s flares', len(flare_indices))

        for start, _, end in flare_indices:
            data_flare[0].extend(data['time'].iloc[start:end])
            data_flare[1].extend(data['flux'].iloc[start:end])
            
        for start, _, end in flare_indices:
            data = data.drop(index=range(start, end))

    afterglow_par, afterglow_err, afterglow_stats, breaknum = find_afterglow_fit(data, data_flare)

    slopes     = list(afterglow_par[:breaknum+1])
    slopes_err = list(afterglow_err[:breaknum+1])
    breaks     = list(afterglow_par[breaknum+1:-1])
    breaks_err = list(afterglow_err[breaknum+1:-1])
    normal     = afterglow_par[-1]
    normal_err = afterglow_err[-1]

    # Calculate fluence.
    afterglow_fluence = calculate_afterglow_fluence(data, breaknum, breaks, afterglow_par, count_ratio)

    return {'params': {
                'break_num': breaknum,
                'slopes': slopes, 'slopes_err': slopes_err,
                'breaks': breaks, 'breaks_err': breaks_err,
                'normal': normal, 'normal_err': normal_err},
            'fluence': afterglow_fluence,
            'fit_statistics': afterglow_stats}

################################################################################
### FIT FLARES
################################################################################

def fitFlares(data, flare_indices, afterglow, *, count_ratio=1.0, flare_model='fred', skip_mcmc=False):

    if not flare_indices:
        return False
    
    flare_model = fred_flare

    # Fit each flare.
    flare_fits, flare_stats, flare_errs, flare_indices = flare_fitter(data, afterglow, flare_indices, model=flare_model)

    # Format, calculate times, fluence.
    flaresDict = package_flares(data, flare_fits, flare_stats, flare_errs, flare_indices, count_ratio=count_ratio)

    return flaresDict

################################################################################
### FIT GRB LIGHTCURVE
################################################################################

def fitGRB(data: pd.DataFrame, *,
           flare_algorithm: str = 'savgol', flare_model: str = 'fred',
           errors_to_std: float = 1.0, count_ratio: float = 1.0):
    # flare_model - use a certain flare model
    # force_breaks - force a certain break_num
    # cont_ratio

    # remove rich_output
    ## TODO ADD DESC HERE
    logger.debug(f"Starting fitGRB")
    if check_data_input(data) == False:
        raise ValueError("check data failed")

    flare_indices = findFlares(data, algorithm=flare_algorithm) # Find flare deviations.
    afterglow = fitAfterglow(data, flare_indices, errors_to_std=errors_to_std, count_ratio=count_ratio) # Fit continuum.
    flares = fitFlares(data, flare_indices, afterglow, count_ratio=count_ratio, flare_model=flare_model) # Fit flares.

    logger.info(f"LAFF run finished.")
    return afterglow, flares

################################################################################
### PLOTTING
################################################################################

def printGRB(data, afterglow, flares):

    print(f"\033[1m// Afterglow - {afterglow['params']['break_num']} breaks\033[0m")

    print("Breaks |", ", ".join(f"{10**x:.0f} \033[2m(-{10**x - 10**(x-dx):.0f},+{10**(x+dx) - 10**x:.0f})\033[0m" for x, dx in zip(afterglow['params']['breaks'], afterglow['params']['breaks_err'])))

    print("Slopes |", ", ".join(f"{x:.2f} \033[2m(+/-{dx:.2f})\033[0m" for x, dx in zip(afterglow['params']['slopes'], afterglow['params']['slopes_err'])))

    print("Normal |", f"{afterglow['params']['normal']:.2f} \033[2m({afterglow['params']['normal_err']:.2f})\033[0m")

    print(afterglow)

    return 

def plotGRB(data, afterglow, flares, show=True, save_path=None, bat=False):
    logger.info(f"Starting plotGRB.")

    plt.rcParams.update({'font.size': 16})

    plt.xlabel("Time (s)")
    plt.ylabel("Flux (units)")

    # Plot lightcurve.
    logger.debug("Plotting lightcurve.")

    plt.errorbar(data.time, data.flux,
                xerr=[-data.time_nerr, data.time_perr], \
                yerr=[-data.flux_nerr, data.flux_perr], \
                marker='', linestyle='None', capsize=0, zorder=1)
    
    if bat == False:
        # Adjustments for xlims, ylims on a log graph.
        upper_flux, lower_flux = data['flux'].max() * 10, data['flux'].min() * 0.1
        plt.ylim(lower_flux, upper_flux)

        lower_time = 0.8 * (data['time'].iloc[0] + data['time_nerr'].iloc[0])
        upper_time = 1.2 * (data['time'].iloc[-1] + data['time_perr'].iloc[-1])
        plt.xlim(lower_time, upper_time)
        plt.loglog()

    # For smooth plotting of fitted functions.
    max, min = np.log10(data['time'].iloc[0] + data['time_nerr'].iloc[0]), np.log10(data['time'].iloc[-1] + data['time_perr'].iloc[-1])
    constant_range = np.logspace(min, max, num=5000)

    # Plot continuum model.
    logger.debug('Plotting continuum model.')
    fittedContinuum = broken_powerlaw(afterglow['params'], constant_range)
    total_model = fittedContinuum
    plt.plot(constant_range, fittedContinuum, color='c')

    # Overlay marked flares.
    if flares is not False:
        logger.debug("Plotting flare indices and models.")

        for flare in flares:
            
            # Plot flare data.
            flare_data = data.iloc[flare['indices'][0]:flare['indices'][2]+1]
            plt.errorbar(flare_data.time, flare_data.flux,
                        xerr=[-flare_data.time_nerr, flare_data.time_perr], \
                        yerr=[-flare_data.flux_nerr, flare_data.flux_perr], \
                        marker='', linestyle='None', capsize=0, color='r', zorder=2)

            flare_model = fred_flare(flare['params'], constant_range)
            total_model += flare_model
            plt.plot(constant_range, fred_flare(flare['params'], constant_range), color='tab:green', linewidth=0.6, zorder=3)

    # Plot total model.
    logger.debug('Plotting total model.')
    plt.plot(constant_range, total_model, color='tab:orange', zorder=5)

    # Plot powerlaw breaks.
    logger.debug('Plotting powerlaw breaks.')
    for x_pos in afterglow['params']['breaks']:
        plt.axvline(x=10**x_pos, color='grey', linestyle='--', linewidth=0.5, zorder=0)

    if save_path:
        plt.savefig(save_path)
    logger.info("Plotting functions done, displaying...")
    if show == True:
        # what's the point ... just don't call the function?
        # unless i can return the plot object somehow?
        plt.show()

        
    return



from .utility import check_data_input 
from .utility import calculate_fit_statistics, calculate_par_err, calculate_fluence
from .utility import PAR_NAMES_CONTINUUM, PAR_NAMES_FLARE, STAT_NAMES_CONTINUUM


import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import savgol_filter, find_peaks
from scipy.ndimage import label
from scipy.optimize import least_squares, fmin_slsqp
from scipy.stats import f
from .modelling import fred_flare, sum_residuals
from .utility import calculate_fit_statistics, calculate_fluence, calculate_par_err

def fitPrompt(data):

    data = data[data['flux'] != 0.0].reset_index(drop=True)

    data, flares = filter_data(data)

    # continuum = find_continuum(data, flares)

    flares = fit_flares(data, flares)

    return {'data': data, 'flares': flares}

################################################################################
# DATA FILTERING
################################################################################

def filter_data(data):

    data['savgol'] = savgol_filter(data['flux'], window_length=31, polyorder=3)
    
    data['negative_noise'] = data['flux'].copy()
    data['negative_noise'] = data['negative_noise'].apply(lambda x: min(x, 0))
    data['negative_noise'] = data['negative_noise'].rolling(window=50, min_periods=1).std() * 2
    data.loc[data['time'] < data['time'].iloc[0] * 0.95, 'negative_noise'] *= 2
    data.loc[data['time'] > data['time'].iloc[-1] * 0.95, 'negative_noise'] *= 2
    data.loc[0, 'negative_noise'] = data['negative_noise'].iloc[1]



    # Filter out negatives.
    filtered_data = data['savgol'].copy()
    filtered_data = filtered_data[filtered_data > 0]
    avg_positive = np.average(filtered_data)
    
    # Filter out big peaks.
    filtered_data = filtered_data[filtered_data < 3 * avg_positive]
    avg_filter = np.average(filtered_data)

    # Calculate residuals.
    data['savgol_residuals'] = data['savgol'] - 3 * avg_filter
    data['savgol_residuals'] = data['savgol_residuals'].apply(lambda x: max(x, 0))
    
    # Calculate rolling std.
    # Cutoff higher peaks.
    modified_flux = data['flux'].copy()
    prev_valid = None
    for i in range(len(modified_flux)):
        if modified_flux[i] > 3 * avg_filter:
            if prev_valid is not None:
                modified_flux[i] = prev_valid
        else:

            prev_valid = modified_flux[i]
    data['moving_std'] = modified_flux.rolling(window=501, min_periods=1).std()
    data.loc[0, 'moving_std'] = data['moving_std'].iloc[1]

    # Gather all flare region indices from residuals.
    labelled_array, num_features = label(data['savgol_residuals'] > 0)
    intial_flare_regions = []

    for i in range(1, num_features+1):
        indices = np.where(labelled_array == i)[0]
        if not indices[0] == indices[-1]:
            intial_flare_regions.append((indices[0], indices[-1]))

    flare_indices = []

    for a, b in intial_flare_regions:
        peak_index = data['savgol'].iloc[a:b].idxmax()
        amplitude = data['savgol'].iloc[peak_index]

        if amplitude < 2 * np.average(data['moving_std'].iloc[a:b]):
            continue
        if len(range(a, b)) <= 1:
            continue

        flare_indices.append((a, b))

    flare_indices = sorted(set(flare_indices))

    return data, flare_indices


################################################################################
# FLARE FITTING
################################################################################

def fit_flares(data, flare_indices):

    flare_data = data.copy()
    flare_data['untouched_flux'] = flare_data['flux']
    flare_data['flux'] -= flare_data['moving_std']
    flare_data['flux'] = flare_data['flux'].apply(lambda x: max(x, 0))

    flares = []

    for dev_start, dev_end in flare_indices:

        try_another = True
        flare_count = 1

        found_maxima, properties = find_peaks(flare_data['flux'].iloc[dev_start:dev_end+1], prominence=flare_data['moving_std'].iloc[dev_start])            
        ranked_maxima_idx = np.argsort(properties['prominences'])[::-1]

        t_start = data['time'].iloc[dev_start]
        t_end   = data['time'].iloc[dev_end]

        while try_another == True:

            input_par = []
            bounds = []

            peaks = [found_maxima[i]+dev_start for i in ranked_maxima_idx[:flare_count]]

            if len(peaks) == 0 and flare_count == 1:
                peaks = [data['savgol'].iloc[dev_start:dev_end].idxmax()]

            if len(peaks) != flare_count:
                try_another = False
                continue

            for i, peak in enumerate(peaks):

                # Parameter guesses
                t_peak    = data['time'].iloc[peak]
                rise      = (t_peak - t_start) / (6 * i+1)
                decay     = (t_end - t_peak)   / (4 * i+1)
                sharp     = 2
                amplitude = data['savgol'].iloc[peak]

                input_par.extend((t_peak, rise, decay, sharp, amplitude))
            
                # Parameter bounds
                t_peak_bound    = [data['time'].iloc[dev_start], data['time'].iloc[dev_end]]
                rise_bound      = [rise/10, rise/3]
                decay_bound     = [decay/10, decay/3]
                sharp_bound     = [1.0, 5.0]
                amplitude_bound = [0, 2 * data['flux'].max()]

                bounds.extend((t_peak_bound, rise_bound, decay_bound, sharp_bound, amplitude_bound))

            def all_constraints():
                pass

            fitted_flare = fmin_slsqp(sum_residuals, input_par, bounds=bounds, args=(flare_data.time, flare_data.untouched_flux, flare_data.flux_perr), iter=200, iprint=0)

            fitted_stats = calculate_fit_statistics(flare_data, fred_flare, fitted_flare)

            if flare_count == 1:
                prev_fits = fitted_flare
                prev_stat = fitted_stats
                flare_count += 1
                continue

            else:
                if fitted_stats['BIC'] + 6 < prev_stat['BIC']:
                    prev_fits = fitted_flare
                    prev_stat = fitted_stats
                    flare_count += 1
                    continue
                else:
                    fitted_flare = prev_fits
                    fitted_stats = prev_stat
                    flare_count -= 1
                    try_another = False

        for i in range(0, len(fitted_flare), 5):

            if fitted_flare[i+4] < 1 * np.average(data['negative_noise'].iloc[dev_start:dev_end]):
                # print('cooked that flare', fitted_flare[i])
                continue

            # if data['time'].iloc[-1]

            flare_data['flux'] -= fred_flare(fitted_flare[i:i+5], flare_data['time'])
            flare_data['flux'] = flare_data['flux'].apply(lambda x: max(x, 0))

            fluence_rise = calculate_fluence(fred_flare, fitted_flare[i:i+5], t_start, fitted_flare[i], 1)
            fluence_decay = calculate_fluence(fred_flare, fitted_flare[i:i+5], fitted_flare[i], t_end, 1)  
            fluence_total = fluence_rise + fluence_decay
            
            flares.append({'indices': (dev_start, dev_end), 'parameters': fitted_flare[i:i+5], 'fluence': [fluence_rise, fluence_decay, fluence_total], 'fit_statistics': fitted_stats})


    return flares

    
################################################################################
# PLOTTING
################################################################################

def plotPrompt(prompt_fit, **kwargs):
    """_summary_

    Args:
        data (pd.DataFrame): _description_
        prompt_fit (_type_): _description_
    
    Kwargs:
        grb_name (str):   grb name string to title the plot.
        zero_lines (bool): plot the y=0 dashed lines.
        main_data (bool): plot the main data, default true
        residuals (bool): plot the residuals, default true
        flare_spans (bool): plot the flare spans, default true.
        flare_fit
        total fit
        save
        TODO savepath
    """

    data = prompt_fit['data']

    residuals = kwargs.get('residuals', True)

    if residuals:
        fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True, figsize=(10,6), gridspec_kw={'hspace': 0})
    else:
        fig, ax1 = plt.subplots(figsize=(10, 6))

    constant_range = np.linspace(data['time'].iloc[0], data['time'].iloc[-1], num=len(data['time']) * 10)

    # Plot title.
    if (grb_name := kwargs.get('grb_name')):
        fig.suptitle(str(grb_name))
    
    # Zero lines.
    if kwargs.get('zero_lines', True):
        ax1.axhline(y=0, linestyle='--', color='grey', linewidth=0.5)
        if residuals:
            ax2.axhline(y=0, linestyle='--', color='grey', linewidth=0.5)

    # Main data points.
    if kwargs.get('main_data', True):
        ax1.errorbar(data['time'], data['flux'], yerr=data['flux_perr'], linestyle='None', marker='.', color='grey', linewidth=0.5, alpha=0.4, zorder=-1)

    # Savgol filter line.
    if kwargs.get('savgol', True):
        ax1.plot(data['time'], data['savgol'], color='black', linewidth=0.5, linestyle='--')

    # Residuals.
    if residuals:
        ax2.plot(data['time'], data['savgol_residuals'], linewidth=0.5, color='tab:green')


    total_model = [0.0] * constant_range

    for flare in prompt_fit['flares']:

        srt, end = flare['indices']

        if kwargs.get('flare_spans', True):
            ax1.axvspan(data['time'].iloc[srt], data['time'].iloc[end], color='b', alpha=0.2)
            if residuals:
                ax2.axvspan(data['time'].iloc[srt], data['time'].iloc[end], color='b', alpha=0.2)

        flare_model = fred_flare(flare['parameters'], constant_range)
        total_model += flare_model

        if kwargs.get('flare_fit', True):
            ax1.plot(constant_range, flare_model, color='#2274A5', linewidth=1)
            if residuals:
                ax2.plot(constant_range, flare_model, color='cyan', linewidth=1)

    if kwargs.get('total_fit', True):
        ax1.plot(constant_range, total_model, color='tab:orange', linewidth=2)

    plt.xlabel('Time since trigger (s)', fontsize=16)
    plt.ylabel('Count rate (counts/s)', fontsize=16)

    # plt.axvline(data['time'].iloc[0] * 0.95, color='r')
    # plt.axvline(data['time'].iloc[-1] * 0.95, color='r')
    # plt.plot(data['time'], data['negative_noise'], color='m')

    # plt.xlim(-50, 200)
    # plt.ylim(-0.2, 0.4)
    if (save_path := kwargs.get('save')):
        plt.savefig(save_path + grb_name + '.png', bbox_inches='tight')
    if kwargs.get('show', True):
        plt.show()


import numpy as np
import pandas as pd
import scipy.integrate as integrate
from scipy.optimize import approx_fprime
import math
import logging

logger = logging.getLogger('laff')

PAR_NAMES_FLARE = ['t_start', 'rise', 'decay', 'amplitude']
PAR_NAMES_CONTINUUM = ['break_num', 'slopes', 'slopes_err', 'breaks', 'breaks_err', 'normal', 'normal_err']
STAT_NAMES_CONTINUUM = ['chisq', 'rchisq', 'n', 'npar', 'dof', 'deltaAIC']

def calculate_fit_statistics(data, model, params, y_col='flux'):
    
    # if temp_flare_shell:
        # fitted_model = model(np.array(data.time), params)
    # else:
        # fitted_model = model(params, np.array(data.time))

    chisq = np.sum(((data[y_col] - model(params, data['time'])) / data['flux_perr']) ** 2)  
    n = len(data['time'])
    npar = len(params)
    
    dof = n - npar
    r_chisq = chisq / dof

    deltaAIC = (2 * npar) + (n * np.log(r_chisq))
    deltaAIC = deltaAIC if deltaAIC != -np.inf else np.inf # negative infinity check

    residuals = data[y_col] - model(params, data['time'])
    variance = data['flux_perr'] ** 2
    lnL = -0.5 * np.sum((residuals**2) / variance + np.log(2 * np.pi * variance))

    BIC = (np.log(n) * npar) - 2 * lnL
        
    # return {'n': n, 'npar': npar, 'dof': dof, 'deltaAIC': deltaAIC}
    return {'chisq': chisq, 'rchisq': r_chisq, 'n': n, 'npar': npar, 'dof': dof, 'deltaAIC': deltaAIC, 'BIC': BIC}

def check_data_input(data):

    # if not isinstance(data, pd.DataFrame):
        # raise TypeError(f"Invalid input data type. Should be pandas dataframe.")

    if len(data) < 3:
        logger.critical("Too short.")
        raise ValueError(f"Data too short, require at least 3 data points - given data is {len(data)}")

    if not data.shape[1] == 4 and not data.shape[1] == 6:
        raise ValueError("Expected dataframe of shape (X, 4) or (X, 6)")

    # Check column names.
    expected_columns = ['time', 'time_perr', 'time_nerr', 'flux', 'flux_perr', 'flux_nerr']
    if data.shape[1] == 4:
        data.columns = ['time', 'time_perr', 'flux', 'flux_perr']
        data['time_nerr'] = data['time_perr']
        data['flux_nerr'] = data['flux_perr']
        data.columns = expected_columns
    elif data.shape[1] == 6:
        data.columns = expected_columns
    else:
        raise ValueError(f"Expected dataframe with 4 or 6 columns - got {data.shape[1]}.")
    
    data = data.reset_index(drop=True)


    logger.debug('Data input is good.')

    return True

def calculate_fluence(model, params, start, stop, count_ratio):
    """Given some model and range, calculate the fluence."""

    range = np.logspace(np.log10(start), np.log10(stop), num=2500)
    fitted_model = model(params, range)
    fluence = integrate.trapezoid(fitted_model, x=range)
    logger.debug('Fluence from %s to %s is %s', start, stop, fluence)

    return fluence * count_ratio

def calculate_par_err(params, chi_wrapper):

    epsilon = [max(x * 1e-4, 1e-8) for x in params]
    n = len(params)
    hessian = np.zeros((n, n))

    for i in range(n):
        # vary
        x1 = np.array(params, copy=True)
        x1[i] = params[i] + epsilon[i]
        grad1 = approx_fprime(x1, chi_wrapper, epsilon)
        x1[i] = params[i] - epsilon[i]
        grad2 = approx_fprime(x1, chi_wrapper, epsilon)
        hessian[:, i] = (grad1 - grad2) / (2 * epsilon[i])

    try:
        covariance_matrix = np.linalg.inv(hessian)
    except np.linalg.LinAlgError:
        covariance_matrix = np.linalg.pinv(hessian) # Moore–Penrose pseudo-inverse
        logger.warning("\tusing pseudo inverse cov matrix - par likely poorly constrained")
    errors = np.sqrt(np.diag(covariance_matrix))

    return 3 * errors


import numpy as np
import logging
from scipy.optimize import fmin_slsqp
from scipy.signal import find_peaks
from scipy.stats import f
from ..utility import calculate_fit_statistics, calculate_par_err, calculate_fluence 
from ..modelling import broken_powerlaw

logger = logging.getLogger('laff')

#################################################################################
### FRED MODEL
#################################################################################

def fred_flare(params, x):
    # J. P. Norris et al., ‘Attributes of Pulses in Long Bright Gamma-Ray Bursts’, The Astrophysical Journal, vol. 459, p. 393, Mar. 1996, doi: 10.1086/176902.
    
    x = np.array(x)

    flr_params = [params[i:i+5] for i in range(0, len(params), 5)]

    total_model = [0.0] * len(x)

    for flr in flr_params:
        t_max, rise, decay, sharpness, amplitude = flr

        flr_model = amplitude * np.exp( -(abs(x - t_max) / rise) ** sharpness)
        flr_model[np.where(x > t_max)] = amplitude * np.exp( -(abs(x[np.where(x > t_max)] - t_max) / decay) ** sharpness)

        total_model += flr_model

    return total_model

def sum_residuals(params, *args):
    x, y, y_err = args
    return np.sum(((y - fred_flare(params, x)) / y_err) ** 2)

#################################################################################
### SCIPY.ODR FITTING
#################################################################################

def flare_fitter(data, continuum, flares, model='fred'):
    """ 
    Flare fitting function. Takes already found flare indices and models them.

    Also runs:
      - 
    
    """

    logger.info("Fitting flares...")

    data['residuals'] = data['flux'] - broken_powerlaw(continuum['params'], data['time'])

    flareFits    = []
    flareStats   = []
    flareErrs    = []
    flareIndices = []

    for start, peak, end in flares:

        try_another = True
        flare_count = 1

        t_start = data['time'].iloc[start]
        t_peak  = data['time'].iloc[peak]
        t_end   = data['time'].iloc[end]

        while try_another == True:

            input_par = []

            peak_guesses = np.linspace(np.log10(t_start), np.log10(t_end), num=flare_count+2)[1:-1]
            peak_guesses = [10**x for x in peak_guesses]

            for i in range(flare_count):
                # Parameter guesses.
                t_max     = peak_guesses[i] if flare_count > 1 else t_peak
                rise      = (t_peak - t_start) / (3 * flare_count)
                decay     = (t_end - t_peak) / (2 * flare_count)
                sharpness = 2
                amplitude = data['residuals'].iloc[peak]

                input_par.extend((t_max, rise, decay, sharpness, amplitude))

            bounds = flare_count * ([t_start, t_end], [rise/10, t_end-t_start], [decay/10, t_end-t_start], [1.0, 3.0], [data['flux'].min(), data['flux'].max()])

            def all_constraints(params, *args):

                temp_start = start
                while data['residuals'].iloc[temp_start] <= data['flux'].iloc[temp_start] * 1e-3:
                    temp_start += 1

                temp_end = end
                while data['residuals'].iloc[temp_end] <= data['flux'].iloc[temp_end] * 1e-3:
                    temp_end -= 1

                # start, peak and end points should be upper limits
                x_points = [data['time'].iloc[x] for start, peak, end in flares for x in (temp_start, peak, end)]
                y_points = [data['residuals'].iloc[x] for start, peak, end in flares for x in (temp_start, peak, end)]
                upper_limits = y_points - fred_flare(params, x_points)

                flare_shape = [(params[2+(i*5)] - params[1+(i*5)]) for i in range(flare_count)]

                # return np.concatenate([upper_limits])
                return np.concatenate([upper_limits, flare_shape])

            fitted_flare = fmin_slsqp(sum_residuals, input_par, bounds=bounds, f_ieqcons=all_constraints, args=(data.time, data.residuals, data.flux_perr), iter=1000, iprint=0)

            fitted_stats = calculate_fit_statistics(data, fred_flare, fitted_flare, y_col='residuals')

            # print('flare_count', flare_count, fitted_stats['BIC'])
            if flare_count == 1:
                prev_fits = fitted_flare
                prev_stat = fitted_stats
                flare_count += 1
                continue
            
            else:
                if fitted_stats['BIC'] + 6 < prev_stat['BIC']:
                    prev_fits = fitted_flare
                    prev_stat = fitted_stats
                    flare_count += 1
                    continue
                else:
                    fitted_flare = prev_fits
                    fitted_stats = prev_stat
                    flare_count -= 1
                    try_another = False

        logger.debug(f"Flare {start}/{peak}/{end} fitted")
        logger.debug('\tconsists of %s flares', flare_count)

        for i in range(0, len(fitted_flare), 5):
        
            individual_par = fitted_flare[i:i+5]
            
            def chi2_wrapper(individual_par):
                return sum_residuals(individual_par, data['time'], data['residuals'], data['flux_perr'])
            individual_errors = calculate_par_err(individual_par, chi2_wrapper)
        
            data['residuals'] = data['residuals'] - fred_flare(individual_par, data.time)

            individual_stats = calculate_fit_statistics(data, fred_flare, fitted_flare)

            flareFits.append(list(individual_par))
            flareErrs.append(list(individual_errors))
            flareStats.append(list(individual_stats.values()))
            flareIndices.append([start, peak, end])

            logger.debug('\tparams\t%s', list(round(x, 2) for x in individual_par))
            logger.debug('\terrors\t%s', list(round(x, 2) for x in individual_errors))

    logger.info("Flare fitting complete for all flares.")
    return flareFits, flareStats, flareErrs, flareIndices

def calculate_p_value(par, data, count, chi_1, dof_1):

    chi_2 = sum_residuals(par, data['time'], data['residuals'], data['flux_perr'])
    dof_2 = len(data['flux']) - len(par)
    
    if count == 1:
        return chi_2, dof_2
    
    F = ((chi_1 - chi_2) / (dof_1 - dof_2)) / (chi_2/dof_2)
    p_value = 1 - f.cdf(F, dof_1-dof_2, dof_2)

    return p_value, chi_2, dof_2

#################################################################################
### NEAT PACKAGING
#################################################################################

def package_flares(data, fits, stats, errs, indices, count_ratio=1.0):

    flaresDict = []

    for idx, fit, stat, err in zip(indices, fits, stats, errs):

        # tmax, rise, decay, sharp, ampl
        # 0     1     2      3      4
        fitted_flare = fred_flare(fit, data['time'])

        start_time = fit[0] - (fit[1] * (-np.log(0.001)) **(1/fit[3]))
        peak_time  = fit[0]
        end_time = fit[0] + (fit[2] * (-np.log(0.001)) **(1/fit[3]))
        times = [start_time, peak_time, end_time]

        fluence_rise  = calculate_fluence(fred_flare, fit, start_time, peak_time, count_ratio=count_ratio)
        fluence_decay = calculate_fluence(fred_flare, fit, peak_time, end_time, count_ratio=count_ratio)
        fluences = [fluence_rise, fluence_decay, fluence_rise + fluence_decay]

        peak_flux = max(fitted_flare)
        
        flaresDict.append({'times': times, 'indices': idx, 'params': fit, 'stats': stat, 'errors': err, 'fluence': fluences, 'peak_flux': peak_flux})

    return flaresDict